# -*- coding: utf-8 -*-
"""
Created on Thu Apr 30 09:10:51 2020
@author: Luja von Köckritz, Boris van Beijnum & Sebastiaan Deetman

@researcher and author report: Boris van Beijnum
"""

""" 
Note by the Author: this Python program is supplementary material (SM1) for the Industrial Ecology graduation project 'Cost of Comfort: waking up to furniture waste', authored by Boris van Beijnum (2021).
It builds on previous modelling efforts by von Köckritz (2020), from here on called the 'Base model', of which the original source code is provided seperatly in the Supplementary material (SM2 - Base model). All parts of the study model
that are adopted from the base model are indicated in the in-code notes. The Dynamic stock model function (inflow_outflow_2ndhand) is developed by Deetman (2021) with contributions by the author.

Model Guide:
The model consists of multiple parts:
Part 1: Preparing data
Part 2: Required stock & defining scenarios
Part 3: DSM function (1st hand stock, 2nd hand stock, 1st- and 2nd hand inflow and outflow)
Part 4: Recycling function (recycled material and landfilled material)
Part 5: Miscellaneaous

Most of the data preparation of part 1 is adopted from the Base model. Part 2 of the model is primarily meant to calculate the 'required stock' (see description in report).
In the third part of the model, the required stock is used in the DSM function together with product lifetimes and a dynamic reuse fraction to calculate 1st and 2nd hand stocks, in- and outflows.
In the fourth part of the model, the outcomes of the DSM function are used to calculate the total amount of landfilled and recycled material [Mt] for Europe.

(!!!) Scenario analysis: when running the program, please select a scenario in the first block of part 1. Scenario '0' = Baseline, scenario '1' = Student scenario

Important Dataframes (Europe only):
stock_europe_kg_mc_areachart                : contains stock in Mt per material category and per year (Note: kg are converted to Mt)
recycling_total_areachart                   : contains total amount of recycled material in Mt per material category and per year
landfill_total_areachart                    : contains total amount of landfilled material in Mt per material category and per year
inflow_1st_areachart                        : contains total amount of material used in primary production in Mt per material category and per year
stock_kg_eu_total                           : contains total stock in Mt divided by 1st and 2nd hand furniture (Note: kg are converted to Mt)
inflow_1st_recyclecorrected_areachart       : contains total amount of required primary material in Mt per material category and per year
stock_europe_kg_total_pc_areachart          : contains stock in Mt per product category and per year (Note: kg are converted to Mt)
stock_europe_units_total                    : contains stock in units per product category

furniture_units_tail                        : required stock per year in units with interpolated historic tail
pop_eu                                      : total European population per year (IMAGE regions: Western Europe + Central Europe)
furniture_weight_raw                        : weight per product category, based on material database
f_amount_raw                                : amount of furniture in units per square meter residential surface
materials_raw                               : material composition per product category
lifetimes                                   : average lifetime per product category
re_use_fraction_peritem                     : reuse fraction per product category
recycle_ratio_1st                           : recycle ratio per product category for 1st hand furniture
recycle_ratio_2nd                           : recycle ratio per product category for 2nd hand furniture
weight_factor_boris                         : weight factor per quintile (see report for description)
relative_reuse_factor                       : relative reuse factor per product category relative to mean reuse factor (see report for description)
items                                       : product category
years                                       : model years (years_tail to include historic tail back to 1900)
"""
#%%
"""-----------------------------------------> Part 1 Preparing data & scenarios <--------------------------------------------------------"""
#%%
# define imports, counters & settings
import pandas as pd
import numpy as np
import os
import matplotlib.pyplot as plt
import scipy
import sys
#import matplotlib.pyplot as plt
#from past.builtins import execfile

path = 'C:\\Users\\Boris\\Documents\\Ausbildung\\Industrial_Ecology\\Thesis\\Python_model\\Furniture_DynamicStockModel_final\\Furniture_DynamicStockModel\\'
#path = 'C:\\Users\\Boris\\Documents\\Ausbildung\\Industrial_Ecology\\Thesis\\Python_model\\Furniture_DynamicStockModel_final\\Furniture_DynamicStockModel\\'
os.chdir(path)
datapath = path + 'Furniture_data\\'
scenario = "SSP2"
variant  = "BL"    
imagepath = path + scenario + "\\" +  scenario + "_"  + variant + "\\"

#execfile('read_mym.py')         # module to read in mym (IMAGE output) files
idx = pd.IndexSlice             # needed for slicing multi-index dataframes (pandas)
scenario = "SSP2"
variant  = "BL"    

startyear = 1971
endyear = 2100
regions = 26
regions_list = ['1','2','3','4','5','6','7','8','9','10','11','12','13','14','15','16','17','18','19','20','21','22','23','24','25','26']
regions_names =  ['Canada','USA','Mexico','Central America', 'Brazil','Rest of South America','Northern Africa','Western Africa','Eastern Africa','South Africa','Western Europe','Central Europe','Turkey','Ukraine region','Central Asia','Russia region','Middle East','India','Korea region','China region','Southeastern Asia','Indonesia region','Japan','Oceania','Rest of South Asia','Rest of Southern Africa']
#regions_dict = dict(zip(regions_list, regions_names))
years = list(range(startyear,endyear+1))
quintiles = ['Q1','Q2','Q3','Q4','Q5']

"""--> Select a scenario here (default is Baseline = 0) <--"""
scenario = 0   #<---------------- Choose the scenario here: 0 = Baseline, 1 = Student scenario (see report for descripton)
#%% Define model functions
"""Note: the 'interpolate_data' and 'stock_tail' functions are adopted from the 'base model' and defined by Luja von Köckritz (2020). The original source code of the Base model (SM2), which is developed by von Köckritz (2020), as well as the original data files used in the Base model, are available as supplementary material"""
# some csv files are defined only for a few years, this function ensures that a linear interpolation us used to derive data for every year between 1971 and 2100
def interpolate_data(data):
    data = data.append(pd.Series(data.loc[data.last_valid_index()], name=endyear))   
    data = data.append(pd.Series(data.loc[data.first_valid_index()], name=startyear))

    for i in range(startyear, endyear):
        if i in data.index:
            pass
        else:
            data = data.append(pd.Series(name=i))

    data = data.sort_index(axis=0).interpolate(axis=0)
    return data

# In order to calculate inflow & outflow smoothly (without peaks for the initial years), we calculate a historic tail to the stock, by adding a 0 value for first year of operation (=1926), then interpolate values towards 1971
# expected input: columns: regions; index: years 
def stock_tail(stock):
    zero_value = [0 for i in range(0,regions)]
    stock_used = pd.DataFrame(stock).reindex_like(stock)
    stock_used.loc[first_year_furniture] = zero_value  # set all regions to 0 in the year of initial operation
    stock_new  = stock_used.reindex(list(range(first_year_furniture,endyear+1))).interpolate()
    return stock_new
#%% Read in files --------------------------------------------------------------------------
"""Note: preparation of the IMAGE data has been adopted from the Base model, developed by Luja von Köckritz (2020)"""
# IMAGE data
# Interpolate population in millions
raw_pop_data = pd.read_csv(imagepath + 'pop.csv', delimiter=';', index_col = 0)
new_pop_data = raw_pop_data.reindex(years).interpolate()
new_pop_data.columns = regions_names

# Import share of rural population per region per year (IMAGE data)
rur_share = pd.read_csv(imagepath + 'rurpop.csv', delimiter=',', index_col = 0)
rur_share.columns = regions_names
rur_share = interpolate_data(rur_share)
rur_share = rur_share.drop(1970, axis=0)

# Calculate rural and urban population in millions
pop_rur = pd.DataFrame().reindex_like(rur_share)
pop_urb = pd.DataFrame().reindex_like(rur_share)
pop_rur = new_pop_data.multiply(rur_share, axis = 'index')
pop_urb = new_pop_data.multiply(1-rur_share, axis = 'index')
pop_rur = pop_rur.stack()
pop_urb = pop_urb.stack()

# Calculate floor space (m2) per capita for each quintile - split into rural and urban floorspace
raw_fli_data = pd.read_csv(imagepath + 'res_Floorspace.csv', delimiter=';', skiprows = 2, index_col = [0,1])
raw_fli_data.drop(['Total', 'Urban', 'Rural'], axis=1, inplace=True)
raw_fli_data.drop(raw_fli_data.loc[idx[:,27],:].index, inplace=True)
raw_fli_data = raw_fli_data.stack()
raw_fli_data = raw_fli_data.unstack(1)
raw_fli_data.columns = regions_names                # Might be better to do this with a dictionary
raw_fli_data=raw_fli_data.stack().unstack(1)
raw_fli_data = raw_fli_data.rename_axis(index=['t','Region'])

# read in additional (non-IMAGE) furniture data files
materials_raw = pd.read_csv(datapath + 'materials_boris.csv', delimiter=';', index_col = [0, 1])       # kg per unit
f_amount_raw = pd.read_csv(datapath + 'furniture_amounts_boris.csv', delimiter=';', index_col = [0])  # unit per capita per m² for each quintile
lifetimes = pd.read_csv(datapath + 'lifetimes_boris.csv', delimiter=';', index_col = [0])         # years
weight_factor_boris = pd.read_csv(datapath + 'weight_factor_boris.csv', delimiter=';', index_col = [0, 1])
relative_reuse_factor = pd.read_csv(datapath + 'relative_reuse_factor.csv', delimiter=';', index_col = [0])
#%% Calculation (average) raw weight per furniture category 'furniture_weight_raw' based on the sum of the material composition & define a list/ index with the furniture items 'items' for later use
"""Note: calculation of the (raw) mean weight per furniture category is adopted from the Base model, developed by Luja von Köckritz (2020)"""
f_amount_raw   = f_amount_raw.reindex(sorted(f_amount_raw.columns), axis=1) # unit per capita per m² for each quintile
furniture_list = f_amount_raw.index.tolist()
f_index        = pd.MultiIndex.from_product([years, regions_names, quintiles], names=['t', 'Region','Quintile'])

# Calculate total weight of materials in furniture (kg per unit)
furniture_weight_raw = materials_raw.sum(axis=0, level=0)
items            = furniture_weight_raw.T.columns #used since the files do not all contain data on the same furniture items any more
#%% Calculation furniture weight = weight per furniture category including weight factor, 'furniture_weight'
"""Note: the mean weight per furniture category, calculated in the previous sector is weighted with the 'weight factor' (weight_factor_boris) per product category (see study report). The weight factor calculation is based on a similar calculation as is used for the base model. Data for the weight factor is gathered by von Köckritz (2020). The weight factor calculation is provided as an Excel workbook in the supplementary material SM5 'Calculation weight factor' """
furniture_weight = pd.DataFrame(index=furniture_weight_raw.index, columns=quintiles)
furniture_weight_raw_pq = pd.concat([furniture_weight_raw]*5, axis = 1)
furniture_weight_raw_pq.columns = quintiles
weight_factor_boris_pq = weight_factor_boris.unstack(1)
weight_factor_boris_pq.columns = quintiles
furniture_weight = furniture_weight_raw_pq * weight_factor_boris_pq

furniture_weight = furniture_weight.reindex(sorted(furniture_weight.columns), axis=1)
furniture_weight = furniture_weight.T #This dataframe now contains the weight per product category per quintile (weighted with the weight factor by quintile)
#%%
"""-----------------------------------------> Part 2 Required stock & defining scenarios <--------------------------------------------------------"""
#%% Calculation of total floorspace in millions of m2 per quintile (floorspace (m2) /cap * population (mil.)) = million m2
"""Note: calculation of total floorspace is adopted from the Base model, developed by Luja von Köckritz (2020)"""
# Total floorspace in millions of m2 per quintile (floorspace (m2) /cap * population (mil.)) = million m2
floorspace = pd.DataFrame(index=raw_fli_data.index, columns=quintiles)  
for quintile in quintiles:
    floorspace.loc[:, quintile] = raw_fli_data.loc[:, 'Urb_'+ quintile].multiply(pop_urb * 0.2, axis = 0) + raw_fli_data.loc[:, 'Rur_'+ quintile].multiply(pop_rur * 0.2, axis = 0)         # million m²                              
floorspace = floorspace.stack()   # million m2

#%% Calculate required stock in units per region and per product step 1: furniture per m2 per quintile (which is the same per quintile), for calculation purposes
f_amount_raw_transposed = f_amount_raw.transpose() 
f_amount_perquintile = pd.concat([f_amount_raw]*5, axis=1)
f_amount_perquintile.columns = quintiles
f_amount_perquintile_transposed = f_amount_perquintile.transpose()

#%% Calculate required stock in units per region and per product step 2:'furniture_units' is the primary dataframe for the DSM
furniture_units = pd.DataFrame(index=f_index, columns=items)
for quintile in quintiles:
    for item in items:
        furniture_units.loc[idx[:,:,quintile],item] = floorspace.loc[idx[:,:,quintile]].mul(f_amount_perquintile_transposed.loc[quintile, item])

#%% Prepare required stock for DSM function step 1: reorder dataframe to fit DSM function
furniture_units = furniture_units.astype('float64') #furniture_units = required stock (without historic tail)
furniture_units = furniture_units.stack()
furniture_units = furniture_units.unstack(1)
furniture_units = furniture_units.reorder_levels([2, 0, 1])
furniture_units = furniture_units.sort_index(level=0)

#%% Prepare required stock for DSM function step 2: Add a historic tail to the furniture_units dataframe to fit the DSM function
first_year_furniture = 1900
stdev_mult = 0.2 # to be set later

years_tail = list(range(first_year_furniture,endyear+1))
furniture_tail_units = pd.DataFrame(index=pd.MultiIndex.from_product([furniture_list, years_tail, quintiles], names=['t', 'Region','Quintile']), columns=regions_names)
furniture_tail_units = furniture_tail_units.astype('float64')
furniture_tail_units = furniture_tail_units.reindex(sorted(furniture_tail_units.columns), axis=1)

for quintile in quintiles:
    for item in items: 
        furniture_tail_units.loc[idx[item,:,quintile],:] = stock_tail(furniture_units.loc[idx[item,:,quintile],:].droplevel(level=[0,2])).to_numpy()

furniture_tail_units.index.set_names(['Item', 't', 'Quintile'], inplace=True) #use this dataframe for the required stock in the DSM function
#%%
"""The next blocks define the scenario analysis: the reuse fraction of furniture that enters the 2nd hand market and recycle fraction for the amount of waste material that is being recycled
Two scenarios are identified: the Baseline scenario and the Student scenario. In the Baseline scenaio, the reuse- and recycle fraction stay constant. In the student scenario, the reuse factor is gradually increased
with 1% per year until it reaches 35% and the recycle fraction is gradually increased with 2% per year until it reaches 50%. """
#%% Define relative reuse factor (Based on the survey: the re_use fraction per furniture product category relative to the mean re_use fraction) per item and per year
relative_reuse_factor_peryear = pd.concat([relative_reuse_factor]*201, axis=1)
relative_reuse_factor_peryear.columns = years_tail
relative_reuse_factor_peryear = relative_reuse_factor_peryear.stack(0)
relative_reuse_factor_peryear = relative_reuse_factor_peryear.to_frame()
relative_reuse_factor_peryear = relative_reuse_factor_peryear.rename(columns = {0:'relative_reuse_factor'})
relative_reuse_factor_peryear.index.set_names(['Item', 't'], inplace=True)
#%% Define reuse and recycle fraction under baseline and student scenario conditions
if scenario == 0: 
    # re_use fraction for Baseline
    re_use_fraction_peritem = pd.DataFrame(index=items, columns=years_tail).unstack(1)
    re_use_fraction_peritem.index.set_names(['t', 'Item'], inplace = True)
    re_use_fraction_peritem = re_use_fraction_peritem.to_frame()
    re_use_fraction_peritem = re_use_fraction_peritem.rename(columns = {0:'re_use_fraction'})
    re_use_fraction_peritem['re_use_fraction']= 0.06 #determines the total average re_use fraction
    re_use_fraction_peritem = re_use_fraction_peritem.reorder_levels(['Item', 't'], axis=0)
    
    relative_reuse_factor_peryear = relative_reuse_factor_peryear.unstack(1) #prepare relative reuse factor dataframe for multiplication
    re_use_fraction_peritem_2 = re_use_fraction_peritem.unstack(1) #prepare a temporary re_use_fraction dataframe of different shape for calculation purposes
    relative_reuse_factor_peryear.columns = years_tail
    re_use_fraction_peritem_2.columns = years_tail
    re_use_fraction_rrf = pd.DataFrame().reindex_like(relative_reuse_factor_peryear)

    # Characterise re_use fraction with relative reuse factor = 're_use_fraction_rrf': Characterise re_use_fraction with relative reuse factor and prepare dataframe for DSM function
    re_use_fraction_rrf = relative_reuse_factor_peryear * re_use_fraction_peritem_2 #Characterise re_use_fraction with relative reuse factor
    re_use_fraction_rrf = re_use_fraction_rrf.stack(0) 
    re_use_fraction_rrf = re_use_fraction_rrf.to_frame()
    re_use_fraction_rrf = re_use_fraction_rrf.rename(columns = {0:'re_use_fraction'}) #Use this dataframe for the dynamic reuse fraction in the DSM function

    # recycle fraction for Baseline
    recycle_ratio = pd.DataFrame(index = re_use_fraction_rrf.index, columns={'recycle_ratio'})
    recycle_ratio_1st = recycle_ratio.fillna(0.1) # Recycle fraction per product category and per year of 10%
    recycle_ratio_2nd = recycle_ratio.fillna(0.1) # Recycle fraction per product category and per year of 10%
    
elif scenario == 1:
    # re_use fraction for Student scenario
    re_use_fraction_peritem = pd.DataFrame(index=items, columns=years_tail).unstack(1)
    re_use_fraction_peritem.index.set_names(['t', 'Item'], inplace = True)
    re_use_fraction_peritem = re_use_fraction_peritem.to_frame()
    re_use_fraction_peritem = re_use_fraction_peritem.rename(columns = {0:'re_use_fraction'})
    re_use_fraction_peritem['re_use_fraction']= 0.06 #determines the total average re_use fraction
    re_use_fraction_peritem = re_use_fraction_peritem.reorder_levels(['Item', 't'], axis=0)
    
    relative_reuse_factor_peryear = relative_reuse_factor_peryear.unstack(1) #prepare relative reuse factor dataframe for multiplication
    re_use_fraction_peritem_2 = re_use_fraction_peritem.unstack(1) #prepare a temporary re_use_fraction dataframe of different shape for calculation purposes
    relative_reuse_factor_peryear.columns = years_tail
    re_use_fraction_peritem_2.columns = years_tail
    re_use_fraction_rrf = pd.DataFrame().reindex_like(relative_reuse_factor_peryear)
    
    # increase reuse fraction with 1% per year until reuse fraction reaches 35%    
    for item in items:
        for step in range(2021,2101):
            n = float(re_use_fraction_peritem_2.loc[item, step-1])
            if (n<0.35):
                re_use_fraction_peritem_2.loc[item, step] = n+float(0.01)
            else:
                re_use_fraction_peritem_2.loc[item, step] = float(0.35)
                
    # Characterise re_use fraction with relative reuse factor = 're_use_fraction_rrf': Characterise re_use_fraction with relative reuse factor and prepare dataframe for DSM function
    re_use_fraction_rrf = relative_reuse_factor_peryear * re_use_fraction_peritem_2 #Characterise re_use_fraction with relative reuse factor
    re_use_fraction_rrf = re_use_fraction_rrf.stack(0) 
    re_use_fraction_rrf = re_use_fraction_rrf.to_frame()
    re_use_fraction_rrf = re_use_fraction_rrf.rename(columns = {0:'re_use_fraction'}) #Use this dataframe for the dynamic reuse fraction in the DSM function
    
    # recycle fraction for Student scenario
    recycle_ratio = pd.DataFrame(index = re_use_fraction_rrf.index, columns={'recycle_ratio'})
    recycle_ratio_1st = recycle_ratio.fillna(0.1) # Recycle fraction per product category and per year of 10%
    recycle_ratio_2nd = recycle_ratio.fillna(0.1) # Recycle fraction per product category and per year of 10%
                
    # increase recycle fraction with 2% per year until recycle fraction reaches 50%
    for item in items:
        for step in range(2021,2101):
            n = float(recycle_ratio_1st.loc[item, step-1])
            if (n<0.5):
                recycle_ratio_1st.loc[item, step] = n+float(0.02)
            else:
                recycle_ratio_1st.loc[item, step] = float(0.5)
                
    for item in items:
        for step in range(2021,2101):
            n = float(recycle_ratio_2nd.loc[item, step-1])
            if (n<0.5):
                recycle_ratio_2nd.loc[item, step] = n+float(0.02)
            else:
                recycle_ratio_2nd.loc[item, step] = float(0.5)

else:
    sys.exit("No (correct) scenario has been chosen. Please choose a scenario at the top of the program")
#%%
"""-----------------------------------------> Part 3 DSM function <--------------------------------------------------------"""

"""Note: the DSM function is developed by (Deetman, 2021) with contributions by the author"""
#%% INFLOW OUTFLOW calculations that distinguish first and second hand stock
"""
This function takes a given in-use stock of products and generates the corresponding 1st and a 2nd hand inflow, stock and outflow, based on a fixed re-use fraction and a product-specific lifetime
Inputs: 1) pd dataframe describing the stock of a single product over time (rows) and by region (columns)
        2) 1 value describing the mean lifetime of that product
        3) pd dataframe describing the fraction of reuse over time (rows) #Boris
Returns:1) 1st hand inflow (by year and region)
        2) 1st hand stock (by cohort, year & region) 
        3) 1st hand outflow (by cohort, year & region)
        4) 2nd hand inflow (by year and region)
        5) 2nd hand stock (by cohort, year & region) 
        6) 2nd hand outflow (by cohort, year & region)
Based on a generic re-use fraction of the 1st hand waste-flow (defined outside the function for now)
""" 

first_year      = 1900                           # currently given, but could be based on stock.first_valid_index() inside the function
re_use_fraction = 0.06                           # make a list in order to change the re-use fraction over time
stdev_mult      = 0.214                          # currently static standard deviation, could also be made dynamic / product-dependent

def inflow_outflow_2ndhand(stock, lifetime, re_use_fraction_b):
   
   lifetime_first  = lifetime
   lifetime_second = lifetime # currently set to match the first-hand lifetime, could be changed with an additonal input to the function
   
   # lifetime assumptions are currently static (one mean & one standard-deviation is applied for every year), but mean lifetimes could change over time if needed 
   mean_first   = lifetime_first
   mean_second  = lifetime_second
   stdev_first  = mean_first * stdev_mult
   stdev_second = mean_second * stdev_mult
   
   # define empty multi-dimensional regional dataframes to be filled & returned (rows: years, columns: regions + years), so years*years because the full cohort matrix is returned
   stock_cohort_1st_regional   = pd.DataFrame(index=list(stock.index), columns=pd.MultiIndex.from_product([list(stock.columns), list(stock.index)]))
   outflow_cohort_1st_regional = pd.DataFrame(index=list(stock.index), columns=pd.MultiIndex.from_product([list(stock.columns), list(stock.index)]))
   inflow_1st_regional         = pd.DataFrame(index=list(stock.index), columns=list(stock.columns))
   
   stock_cohort_2nd_regional   = pd.DataFrame(index=list(stock.index), columns=pd.MultiIndex.from_product([list(stock.columns), list(stock.index)]))
   outflow_cohort_2nd_regional = pd.DataFrame(index=list(stock.index), columns=pd.MultiIndex.from_product([list(stock.columns), list(stock.index)]))
   inflow_2nd_regional         = pd.DataFrame(index=list(stock.index), columns=list(stock.columns))
   
   for region in list(stock.columns):
      
      # before looping over every year, first define the required empty dataframes for both first & second hand
      stock_first_hand    = pd.DataFrame(index=list(stock.index), columns =list(stock.index))
      inflow_first_hand   = pd.Series(index=list(stock.index))
      outflow_first_hand  = pd.DataFrame(index=list(stock.index), columns =list(stock.index))
      
      stock_second_hand   = pd.DataFrame(index=list(stock.index), columns =list(stock.index))
      inflow_second_hand  = pd.Series(index=list(stock.index))
      outflow_second_hand = pd.DataFrame(index=list(stock.index), columns =list(stock.index))
      
      # then define the survival function based on a folded-normal distribution (mind, for a changing lifetime, this needs to be part of the yearly-loop)
      length = len(list(stock.index))
      survival_init_first  = scipy.stats.foldnorm.sf(np.arange(0,length), mean_first/stdev_first,  0, scale=stdev_first)
      survival_init_second = scipy.stats.foldnorm.sf(np.arange(0,length), mean_first/stdev_second, 0, scale=stdev_second)
      
      """
      Loop over every year and fill the stock, inflow & outflow dataframes based on data from previous years (mind stock & outflow are by age-cohort, so dimensions are time*time - i.e. the stock in year z is the remaining inflow from year a,b,c etc.)
      """
      for year in list(stock.index):
         
         if year == first_year:
            # initiate a first year (without knowledge of the years before, assuming this is the start of the stock buildup and both outflow and second hand stock are 0)
            inflow_first_hand[first_year]          = stock[region][first_year]
            outflow_first_hand.loc[:,first_year]   = pd.Series(0,index=stock.index)
            stock_first_hand.loc[:,first_year]     = inflow_first_hand[first_year] * survival_init_first
            
            inflow_second_hand[first_year]         = 0
            outflow_second_hand.loc[:, first_year] = pd.Series(0,index=stock.index)
            stock_second_hand.loc[:, first_year]   = pd.Series(0,index=stock.index)
            
         else:
            # for other years, determine the stock, inflow & outflow year by year
            outflow_first_hand.loc[year:, year] = stock_first_hand.loc[year-1:,year-1].shift(1, axis=0) - stock_first_hand.loc[year-1:,year-1]
            inflow_second_hand[year]           = outflow_first_hand.loc[year,:].sum() * re_use_fraction_b.loc[year]
            stock_second_hand.loc[year:, year] = inflow_second_hand[year] * survival_init_second[0:length-(year-first_year)]             # the survival function gets shortened by 1 year every year
            outflow_second_hand.loc[:, year]   = stock_second_hand.loc[:,year-1].shift(1, axis=0) - stock_second_hand.loc[:,year-1]
            # first hand inflow is defined as the required stock minus the first hand stock that remains in use from previous years, minus the second hand stock
            if (stock[region][year] - stock_second_hand.loc[year,:year].sum() - stock_first_hand.loc[year,:year-1].sum() > 0):
                inflow_first_hand[year] = stock[region][year] - stock_second_hand.loc[year,:year].sum() - stock_first_hand.loc[year,:year-1].sum()
                stock_first_hand.loc[year:,year]   = inflow_first_hand[year] * survival_init_first[0:length-(year-first_year)]
            else:
                inflow_first_hand[year] = 0
                negative_stock = stock[region][year] - stock_second_hand.loc[year,:year].sum() - stock_first_hand.loc[year,:year-1].sum()
                stock_correction_total = negative_stock*int(-1) # negative inflow is 'surplus' which should be corrected in the stock by adding it to the outflow
                stock_correction_cohort = stock_first_hand.loc[year,:year-1] / stock_first_hand.loc[year,:year-1].sum()
              
                # Regular survival of inflow (even though inflow = 0) 
                stock_first_hand.loc[year:,year]   = inflow_first_hand[year] * survival_init_first[0:length-(year-first_year)]
              
                # Actual correction of stock (natural outflow is known & pre-calculated, so only extra correction of negative inflow needs to be subtracted, based on relative share of the age cohorts in total stock)
                for year_correct in range(year, endyear):
                    stock_first_hand.loc[year_correct,:year-1] = stock_first_hand.loc[year_correct,:year-1] - (stock_correction_cohort * stock_correction_total)
                
                # After stock correction, the stock from old age cohorts might become negative. To correct this: set all NaN valus to 0, and then set negative stock to 0
                stock_first_hand = stock_first_hand.fillna(0)
                stock_first_hand[stock_first_hand < 0] = 0
                
                # Actual correction of outflow    
                outflow_first_hand.loc[year,:year-1] = outflow_first_hand.loc[year,:year-1] + (stock_correction_cohort * stock_correction_total) #add stock correction ('surplus') to outflow
                inflow_first_hand.loc[:year-1] = inflow_first_hand.loc[:year-1] - (stock_correction_cohort * stock_correction_total) # (!) CHECK WITH SEBASTIAAN
                
                inflow_second_hand[year]           = outflow_first_hand.loc[year,:].sum() * re_use_fraction_b.loc[year] #stock_correction is now added to outflow, thus increasing 2nd hand inflow
                stock_second_hand.loc[year:, year] = inflow_second_hand[year] * survival_init_second[0:length-(year-first_year)]             # the survival function gets shortened by 1 year every year
                outflow_second_hand.loc[:, year]   = stock_second_hand.loc[:,year-1].shift(1, axis=0) - stock_second_hand.loc[:,year-1]
      # fill the region specific outcomes in an overarching dataframe ()  
      stock_cohort_1st_regional.loc[:,idx[region,:]]   = stock_first_hand.values
      outflow_cohort_1st_regional.loc[:,idx[region,:]] = outflow_first_hand.values
      inflow_1st_regional.loc[:, region]               = inflow_first_hand.values
   
      stock_cohort_2nd_regional.loc[:,idx[region,:]]   = stock_second_hand.values
      outflow_cohort_2nd_regional.loc[:,idx[region,:]] = outflow_second_hand.values
      inflow_2nd_regional.loc[:, region]               = inflow_second_hand.values   

   return inflow_1st_regional, stock_cohort_1st_regional.sum(axis=1, level=0), outflow_cohort_1st_regional.sum(axis=1, level=0), inflow_2nd_regional, stock_cohort_2nd_regional.sum(axis=1, level=0), outflow_cohort_2nd_regional.sum(axis=1, level=0)
#%% Calculate inflow- outflow for one product category and one quintile using the DSM function and the furniture_tail_units dataframe
"""Note: (Deetman, 2021) inactive by default. To test run the DSM function for 1 quintile and 1 product category"""
# define inflow and outflow according to furniture_units (=without historic tail) 
#furniture.drop('Total_Europe', axis=1, inplace=True) #Drop total column to make a dataframe of right length

# IF you 
#index = pd.MultiIndex.from_product([list(furniture_units.index.levels[0]), list(furniture_units.index.levels[1]), list(furniture_units.index.levels[1]), list(furniture_units.index.levels[2])], names=['Item','t','cohorts','Quitile'])

#inflow_units_1st_test  = pd.DataFrame().reindex_like(furniture_units)
#stock_units_1st_test   = pd.DataFrame().reindex_like(furniture_units)
#outflow_units_1st_test = pd.DataFrame().reindex_like(furniture_units)
#inflow_units_2nd_test  = pd.DataFrame().reindex_like(furniture_units)
#stock_units_2nd_test   = pd.DataFrame().reindex_like(furniture_units)
#outflow_units_2nd_test = pd.DataFrame().reindex_like(furniture_units)

# test run on 1 quintile & 1 item (all regions)
#quintile = 'Q5'
#item = 'Armchair'
#results = inflow_outflow_2ndhand(furniture_tail_units.loc[idx[item,:,quintile],:].droplevel(level=[0,2]), lifetimes.loc[item].values[0], re_use_fraction_rrf.loc[idx[item, :], :].droplevel(0).squeeze())
#inflow_units_1st_test.loc[idx[item,:,quintile],:]  = results[0].loc[1971:2100, :].to_numpy()
#stock_units_1st_test.loc[idx[item,:,quintile],:]   = results[1].loc[1971:2100, :].to_numpy()
#outflow_units_1st_test.loc[idx[item,:,quintile],:] = results[2].loc[1971:2100, :].to_numpy()
#inflow_units_2nd_test.loc[idx[item,:,quintile],:]  = results[3].loc[1971:2100, :].to_numpy()
#stock_units_2nd_test.loc[idx[item,:,quintile],:]   = results[4].loc[1971:2100, :].to_numpy()
#outflow_units_2nd_test.loc[idx[item,:,quintile],:] = results[5].loc[1971:2100, :].to_numpy()  

#%%check to see if the given test-run stock (in units), as the driver of the stock model equals the sum of the 1st & second stock
"""Note: (Deetman, 2021) inactive by default. To check if the DSM results are accurate: required stock [i] = 1st hand stock [i] + second hand stock [i]"""
#stock_total_1st_test = stock_units_1st_test.loc[idx[item,:,quintile], 'Canada']
#stock_total_2nd_test = stock_units_2nd_test.loc[idx[item,:,quintile], 'Canada']
#stock_total = stock_total_1st_test + stock_total_2nd_test

#hecksum = stock_total - furniture_tail_units['Canada']
#if checksum.sum() > 0.000001 or checksum.sum() < -0.000001:
#   ctypes.windll.user32.MessageBoxW(0, "Mind, the given stock is not exactly the sum of the 1st and second hand", "Warning", 1)

#%% 
#Apply the dynamic stock model to derive inflow & outflow in units for all items and quintiles
"""Note: (Deetman, 2021) Apply the DSM function for every product category and quintile. Run time for 14 product categories (+/- 2h)"""
inflow_units_1st  = pd.DataFrame().reindex_like(furniture_units)
stock_units_1st   = pd.DataFrame().reindex_like(furniture_units)
outflow_units_1st = pd.DataFrame().reindex_like(furniture_units)
inflow_units_2nd  = pd.DataFrame().reindex_like(furniture_units)
stock_units_2nd   = pd.DataFrame().reindex_like(furniture_units)
outflow_units_2nd = pd.DataFrame().reindex_like(furniture_units)

for quintile in quintiles: # These dataframes contain the results from the DSM function in units
    for item in items:
        results = inflow_outflow_2ndhand(furniture_tail_units.loc[idx[item,:,quintile],:].droplevel(level=[0,2]), lifetimes.loc[item].values[0], re_use_fraction_rrf.loc[idx[item, :], :].droplevel(0).squeeze()) #The DSM function takes one furniture product category per run, meaning that the input should be specified per item (and quintile). 
        inflow_units_1st.loc[idx[item,:,quintile],:]  = results[0].loc[1971:2100, :].to_numpy()
        stock_units_1st.loc[idx[item,:,quintile],:]   = results[1].loc[1971:2100, :].to_numpy()
        outflow_units_1st.loc[idx[item,:,quintile],:] = results[2].loc[1971:2100, :].to_numpy()
        inflow_units_2nd.loc[idx[item,:,quintile],:]  = results[3].loc[1971:2100, :].to_numpy()
        stock_units_2nd.loc[idx[item,:,quintile],:]   = results[4].loc[1971:2100, :].to_numpy()
        outflow_units_2nd.loc[idx[item,:,quintile],:] = results[5].loc[1971:2100, :].to_numpy()        
#%% Save all results to csv files
"""""Note: inactive by default. Use this to save the DSM model function results. These can be read in (next block) for further processing in the recycling function, to save program running time for the necessary DSM function calculations."""
#inflow_units_1st.to_csv('inflow_units_1st_Baseline_check.csv')
#stock_units_1st.to_csv('stock_units_1st_Baseline_check.csv')
#outflow_units_1st.to_csv('outflow_units_1st_Baseline_check.csv')
#inflow_units_2nd.to_csv('inflow_units_2nd_Baseline_check.csv')
#stock_units_2nd.to_csv('stock_units_2nd_Baseline_check.csv')
#outflow_units_2nd.to_csv('outflow_units_2nd_Baseline_check.csv')
#%% Read in csv files containing all the in- and outflow and stock data for all regions 1st and 2nd hand
"""Note: inactive by default. Use this to read-in the data files containing the results from the DSM function for further processing."""
#inflow_units_1st = pd.read_csv('inflow_units_1st_Baseline_check.csv', index_col = [0, 1, 2])
#stock_units_1st = pd.read_csv('stock_units_1st_Baseline_check.csv', index_col = [0, 1, 2])
#outflow_units_1st = pd.read_csv('outflow_units_1st_Baseline_check.csv', index_col = [0, 1, 2])
#inflow_units_2nd = pd.read_csv('inflow_units_2nd_Baseline_check.csv', index_col = [0, 1, 2])
#stock_units_2nd = pd.read_csv('stock_units_2nd_Baseline_check.csv', index_col = [0, 1, 2])
#outflow_units_2nd = pd.read_csv('outflow_units_2nd_Baseline_check.csv', index_col = [0, 1, 2])
#%%
"""-----------------------------------------> Part 3 Recycling function <--------------------------------------------------------"""

#%%
"""Recycling function: can be used to calculate the amount of recycled- and landfilled material from the outflow (for Europe specifically). It takes three dataframe inputs: the (out)flow dataframe (idx[Item, Material, t], quintiles), a dynamic recycle_fraction dataframe (idx[Item, t], recycle_ratio) and dynamic re_use dataframe (idx[Item, t], re_use_fraction), the latter should be the same used in the DSM function. 
The Re-Use fraction dataframe is only necessary for the 1st hand outflow, and contains the fraction of the outflow that goes to the 2nd hand market. For the 2nd hand outflow, this input can be neglected (default = None). NOTE: Important preparation for the recycling function is that the total outflow per year needs to be expressed in the model material categories."""

def recycling_function(flow_dataframe, recycling_ratio, re_use_dataframe=None):
    items_t = list(flow_dataframe.groupby(['Item'], axis=0).sum().index)
    #materials_t = list(flow_dataframe.groupby(['Material'], axis=0).sum().index) #in case you want to specify the recycling ratio to the material category (instead of the product category)
    years_t = list(flow_dataframe.groupby(['t'], axis=0).sum().index)
    quintiles_t = list(flow_dataframe.columns)
    
    recycle_ratio = pd.concat([recycling_ratio.loc[idx[:, 1971:2100], :]]*5, axis=1)
    recycle_ratio.columns = quintiles_t
    landfill_ratio = int(1)-recycle_ratio
    
    recycled_material = pd.DataFrame().reindex_like(flow_dataframe)
    landfilled_material = pd.DataFrame().reindex_like(flow_dataframe)
    
    if re_use_dataframe is not None: #if a re_use dataframe is given for 1st hand outflow, the outflow to waste is determined first based on the re_use fraction
        
        waste = pd.DataFrame().reindex_like(flow_dataframe)
        re_use_t = pd.concat([re_use_dataframe.loc[idx[:, 1971:2100], :]]*5, axis=1) #get rid of historic tail in re_use_fraction dataframe
        re_use_t.columns = quintiles_t
        waste_fraction = int(1)-re_use_t #1-re_use_fraction = waste fraction
        
        for item in items_t: #determine the amount of waste(material), that does not go to 2nd hand market
            for quintile in quintiles_t:
                for year in years_t:
                    waste.loc[idx[item, :, year], quintile] = flow_dataframe.loc[idx[item, :, year], quintile].mul(waste_fraction.loc[idx[item, year], quintile])
                    
        for item in items_t: #waste * recycle_ratio = recycled material
            for quintile in quintiles_t:
                for year in years_t:
                    recycled_material.loc[idx[item, :, year], quintile] = waste.loc[idx[item, :, year], quintile].mul(recycle_ratio.loc[idx[item, year], quintile])    
        
        for item in items_t: #waste * (1-recycle_ratio) = landfilled material
            for quintile in quintiles_t:
                for year in years_t:
                    landfilled_material.loc[idx[item, :, year], quintile] = waste.loc[idx[item, :, year], quintile].mul(landfill_ratio.loc[idx[item, year], quintile])
    else:
        for item in items_t: #waste * recycle_ratio = recycled material
            for quintile in quintiles_t:
                for year in years_t:
                    recycled_material.loc[idx[item, :, year], quintile] = flow_dataframe.loc[idx[item, :, year], quintile].mul(recycle_ratio.loc[idx[item, year], quintile])    
        
        for item in items_t: #waste * (1-recycle_ratio) = landfilled material
            for quintile in quintiles_t:
                for year in years_t:
                    landfilled_material.loc[idx[item, :, year], quintile] = flow_dataframe.loc[idx[item, :, year], quintile].mul(landfill_ratio.loc[idx[item, year], quintile])
    
    return recycled_material, landfilled_material
#%% Calculate in- and outflow and stocks in million kg: multiplied with the weight per units including weight factor per quintile (furniture_weight)
stock_1st_kg = pd.DataFrame().reindex_like(stock_units_1st)
stock_2nd_kg = pd.DataFrame().reindex_like(stock_units_2nd)
outflow_1st_kg = pd.DataFrame().reindex_like(outflow_units_1st)
outflow_2nd_kg = pd.DataFrame().reindex_like(outflow_units_2nd)

for quintile in quintiles:
    for item in items:
        stock_1st_kg.loc[idx[item,:,quintile], :] = stock_units_1st.loc[idx[item,:,quintile], :] *furniture_weight.loc[idx[quintile],item]
        stock_2nd_kg.loc[idx[item,:,quintile], :] = stock_units_2nd.loc[idx[item,:,quintile], :] *furniture_weight.loc[idx[quintile],item]
        outflow_1st_kg.loc[idx[item,:,quintile], :] = outflow_units_1st.loc[idx[item,:,quintile], :] *furniture_weight.loc[idx[quintile],item]
        outflow_2nd_kg.loc[idx[item,:,quintile], :] = outflow_units_2nd.loc[idx[item,:,quintile], :] *furniture_weight.loc[idx[quintile],item]   
#%% Calculate total (= 1st and 2nd hand stock) Europe in units
stock_units_2nd['Europe_stock_2nd'] = stock_units_2nd['Central Europe']+stock_units_2nd['Western Europe']
stock_units_1st['Europe_stock_1st'] = stock_units_1st['Central Europe']+stock_units_1st['Western Europe']
stock_europe_total = pd.DataFrame(index=stock_units_2nd.index, columns=['Europe_stock_1st', 'Europe_stock_2nd'])
stock_europe_total['Europe_stock_1st'] = stock_units_1st['Europe_stock_1st']
stock_europe_total['Europe_stock_2nd'] = stock_units_2nd['Europe_stock_2nd']
stock_europe_units_total = stock_units_1st['Europe_stock_1st'] + stock_units_2nd['Europe_stock_2nd']
stock_europe_units_total = stock_europe_units_total.to_frame()
stock_europe_units_total = stock_europe_units_total.rename(columns = {0:'total_units'})
#%% Calculate total (= 1st and 2nd hand stock) Europe in kg
stock_kg_total_2nd = stock_2nd_kg['Central Europe']+stock_2nd_kg['Western Europe']
stock_kg_total_1st = stock_1st_kg['Central Europe']+stock_1st_kg['Western Europe']
stock_europe_kg_total_pc = stock_kg_total_1st + stock_kg_total_2nd
stock_europe_kg_total_pc = stock_europe_kg_total_pc.to_frame()
stock_europe_kg_total_pc = stock_europe_kg_total_pc.rename(columns = {0:'total_kg'})
stock_europe_kg_total_pc = stock_europe_kg_total_pc.groupby(['Item', 't']).sum()
stock_europe_kg_total_pc = stock_europe_kg_total_pc.unstack(1)
stock_europe_kg_total_pc.columns = years
stock_europe_kg_total_pc_areachart = stock_europe_kg_total_pc.T
#%% Barchart stock step 1: Prepare barchart/ areachart dataframes for specific furniture category stocks
stock_europe_barchart = stock_europe_total.groupby(['Item', 't'], axis=0).sum()
stock_europe_barchart_officechair = stock_europe_barchart.loc[idx['Office chair', :, :]]
stock_europe_barchart_officechair.index = stock_europe_barchart_officechair.index.droplevel(0)
stock_europe_barchart_mattress = stock_europe_barchart.loc[idx['Mattress', :, :]]
stock_europe_barchart_mattress.index = stock_europe_barchart_mattress.index.droplevel(0)
#%% Recycle fraction dataframe: the fraction of the outflow that is being recycled
recycle_fraction = pd.DataFrame(index=re_use_fraction_peritem.index, columns={'recycle_fraction'})
recycle_fraction['recycle_fraction'] = 0.1
#%% Calculate the outflow of 1st and 2nd hand stock per material category step 1: make a datrame which containts the material compositions of the products per year 'Material stock dataframe': for calculation purposes
material_stock_dataframe = pd.concat([materials_raw]*130, axis=1)
material_stock_dataframe.columns = years
material_stock_dataframe_transposed = material_stock_dataframe.transpose()
material_stock_dataframe_transposed.index = years 
material_stock_dataframe_perregion = pd.concat([material_stock_dataframe.stack()]*26, axis=1)
#%% Calculate the outflow of 1st and 2nd hand stock per material category step 2: make a dataframe which contains the outflow in units, which is weighted with the weight factor per quintile for calculation purposes:
weight_factor_boris_peryear = pd.concat([weight_factor_boris]*130, axis=1) #make a dataframe which contains the weight factor indexed by year (for calculation purposes)
weight_factor_boris_peryear.columns = years

outflow_units_1st_eu = pd.DataFrame(index= outflow_units_1st.index, columns={'Outflow'}) #Make a dataframe which contains the outflow for EU specifically
outflow_units_1st_eu['Outflow']=outflow_units_1st['Central Europe'] + outflow_units_1st['Western Europe']
outflow_units_2nd_eu = pd.DataFrame(index= outflow_units_1st.index, columns={'Outflow'}) #Make a dataframe which contains the outflow for EU specifically
outflow_units_2nd_eu['Outflow']=outflow_units_2nd['Central Europe'] + outflow_units_2nd['Western Europe']
inflow_units_1st_eu = pd.DataFrame(index= outflow_units_1st.index, columns={'Inflow'})
inflow_units_1st_eu['Inflow']=inflow_units_1st['Central Europe'] + inflow_units_1st['Western Europe']

outflow_units_1st_eu_wf = pd.DataFrame().reindex_like(outflow_units_1st_eu) #Define dataframe to characterise outflow in units EU with weight factor  (for calculation purposes)
outflow_units_2nd_eu_wf = pd.DataFrame().reindex_like(outflow_units_2nd_eu)
inflow_units_1st_eu_wf = pd.DataFrame().reindex_like(inflow_units_1st_eu)
stock_europe_units_total_wf = pd.DataFrame().reindex_like(stock_europe_units_total)

material_composition_pp_pq = pd.concat([material_stock_dataframe.stack()]*5, axis=1) #define a dataframe which contains the material composition per product item indexed by quintiles (for calculation purposes)
material_composition_pp_pq.columns = quintiles
#%% Calculate the outflow of 1st and 2nd hand stock per material category step 3:First characterise the inflow and outflow in units per quintile with the weight factor
for item in items:
    for quintile in quintiles:
        for year in years:
            outflow_units_1st_eu_wf.loc[item, year, quintile] = outflow_units_1st_eu.loc[item, year, quintile]*weight_factor_boris_peryear.loc[idx[item, quintile], year]
            outflow_units_2nd_eu_wf.loc[item, year, quintile] = outflow_units_2nd_eu.loc[item, year, quintile]*weight_factor_boris_peryear.loc[idx[item, quintile], year]
            inflow_units_1st_eu_wf.loc[item, year, quintile] = inflow_units_1st_eu.loc[item, year, quintile]*weight_factor_boris_peryear.loc[idx[item, quintile], year]
            stock_europe_units_total_wf.loc[item, year, quintile] = stock_europe_units_total.loc[item, year, quintile]*weight_factor_boris_peryear.loc[idx[item, quintile], year]
#%% Calculate the outflow of 1st and 2nd hand stock per material category step 4: Calculate the outflow of 1st and 2nd hand stock per material category
outflow_units_1st_eu_wf_2 = outflow_units_1st_eu_wf.unstack(2) #Change dataframe lay-out to match other dataframes for calculation purposes
outflow_units_1st_eu_wf_2.columns = quintiles
outflow_units_2nd_eu_wf_2 = outflow_units_2nd_eu_wf.unstack(2)
outflow_units_2nd_eu_wf_2.columns = quintiles
inflow_units_1st_eu_wf_2 = inflow_units_1st_eu_wf.unstack(2) #Change dataframe lay-out to match other dataframes for calculation purposes
inflow_units_1st_eu_wf_2.columns = quintiles
stock_europe_units_total_wf_2 = stock_europe_units_total_wf.unstack(2)
stock_europe_units_total_wf_2.columns = quintiles
#%% Calculate the outflow of 1st and 2nd hand stock per material category step 5:
#define the right dataframe like material_composition_pp_pq (pp = per product category, pq = per quintile)
outflow_1st_kg_mc = pd.DataFrame().reindex_like(material_composition_pp_pq)
outflow_2nd_kg_mc = pd.DataFrame().reindex_like(material_composition_pp_pq)
inflow_1st_kg_mc = pd.DataFrame().reindex_like(material_composition_pp_pq)
stock_europe_total_kg_mc = pd.DataFrame().reindex_like(material_composition_pp_pq)

for year in years: #actually calculate the outflow per product- and material category per year
    for item in items:
        for quintile in quintiles:
            outflow_1st_kg_mc.loc[idx[item,:, year], quintile] = material_stock_dataframe.loc[idx[item, :], year]*outflow_units_1st_eu_wf_2.loc[idx[item, year], quintile]
            outflow_2nd_kg_mc.loc[idx[item,:, year], quintile] = material_stock_dataframe.loc[idx[item, :], year]*outflow_units_2nd_eu_wf_2.loc[idx[item, year], quintile]
            inflow_1st_kg_mc.loc[idx[item,:, year], quintile] = material_stock_dataframe.loc[idx[item, :], year]*inflow_units_1st_eu_wf_2.loc[idx[item, year], quintile]
            stock_europe_total_kg_mc.loc[idx[item,:, year], quintile] = material_stock_dataframe.loc[idx[item, :], year]*stock_europe_units_total_wf_2.loc[idx[item, year], quintile]
outflow_1st_kg_mc.index.set_names(['Item', 'Material', 't'], inplace=True)
outflow_2nd_kg_mc.index.set_names(['Item', 'Material', 't'], inplace=True)
inflow_1st_kg_mc.index.set_names(['Item', 'Material', 't'], inplace=True)
stock_europe_total_kg_mc.index.set_names(['Item', 'Material', 't'], inplace=True)
#%% Define 'total' dataframes, including both 1st and 2nd hand outflow, containing the outflow per material category per year
outflow_2nd_kg_mc_total = pd.DataFrame(index=outflow_1st_kg_mc.index, columns={'Outflow'})
outflow_2nd_kg_mc_total['Outflow']=outflow_2nd_kg_mc.sum(axis=1)
outflow_1st_kg_mc_total = pd.DataFrame(index=outflow_1st_kg_mc.index, columns={'Outflow'})
outflow_1st_kg_mc_total['Outflow']=outflow_1st_kg_mc.sum(axis=1)
inflow_1st_kg_mc_total = pd.DataFrame(index=outflow_1st_kg_mc.index, columns={'Inflow'})
inflow_1st_kg_mc_total['Outflow']=inflow_1st_kg_mc.sum(axis=1)
stock_europe_total_kg_mc_total = pd.DataFrame(index=outflow_1st_kg_mc.index, columns={'stock'})
stock_europe_total_kg_mc_total['stock']=stock_europe_total_kg_mc.sum(axis=1)

outflow_1st_kg_mc_total.index.set_names(['Item', 'Material', 't'], inplace=True)
outflow_2nd_kg_mc_total.index.set_names(['Item', 'Material', 't'], inplace=True)
inflow_1st_kg_mc_total.index.set_names(['Item', 'Material', 't'], inplace=True)
stock_europe_total_kg_mc_total.index.set_names(['Item', 'Material', 't'], inplace=True)

outflow_1st_kg_mc_total = outflow_1st_kg_mc_total.groupby(['Material', 't'], axis=0).sum()
outflow_2nd_kg_mc_total = outflow_2nd_kg_mc_total.groupby(['Material', 't'], axis=0).sum()
inflow_1st_kg_mc_total = inflow_1st_kg_mc_total.groupby(['Material', 't'], axis=0).sum()
stock_europe_total_kg_mc_total = stock_europe_total_kg_mc_total.groupby(['Material', 't'], axis=0).sum()

outflow_1st_kg_mc_total = outflow_1st_kg_mc_total.unstack(1)
outflow_1st_kg_mc_total.columns = years
outflow_2nd_kg_mc_total = outflow_2nd_kg_mc_total.unstack(1)
outflow_2nd_kg_mc_total.columns = years
inflow_1st_kg_mc_total = inflow_1st_kg_mc_total.unstack(1)
inflow_1st_kg_mc_total.columns = years
stock_europe_total_kg_mc_total = stock_europe_total_kg_mc_total.unstack(1)
stock_europe_total_kg_mc_total.columns = years #This dataframe contains the total European outflow per year in million kg per material category.
#%% Use the recycle function to determine the amount of recycled and landfilled material from the 1st hand and 2nd hand outflow
outflow_1st_recycle_landfill = recycling_function(outflow_1st_kg_mc, recycle_ratio_1st, re_use_fraction_rrf)
recycling_1st = (outflow_1st_recycle_landfill[0].groupby(['Material','t'], axis=0).sum()).sum(axis=1)
landfill_1st = (outflow_1st_recycle_landfill[1].groupby(['Material','t'], axis=0).sum()).sum(axis=1)

recycling_1st = recycling_1st.to_frame()
recycling_1st = recycling_1st.rename(columns = {0:'recycled_material'})
landfill_1st = landfill_1st.to_frame()
landfill_1st = landfill_1st.rename(columns = {0:'landfilled_material'})

outflow_2nd_recycle_landfill = recycling_function(outflow_2nd_kg_mc, recycle_ratio_2nd)
recycling_2nd = (outflow_2nd_recycle_landfill[0].groupby(['Material','t'], axis=0).sum()).sum(axis=1)
landfill_2nd = (outflow_2nd_recycle_landfill[1].groupby(['Material','t'], axis=0).sum()).sum(axis=1)

recycling_2nd = recycling_2nd.to_frame()
recycling_2nd = recycling_2nd.rename(columns = {0:'recycled_material'})
landfill_2nd = landfill_2nd.to_frame()
landfill_2nd = landfill_2nd.rename(columns = {0:'landfilled_material'})
#%% Area charts of total (1st and 2nd hand) recycling, landfill and nrequired primary material (inflow_1st_recyclecorrected) for Europe
recycling_1st_areachart = recycling_1st.unstack(1)
recycling_1st_areachart.columns = years
recycling_1st_areachart = recycling_1st_areachart.T
recycling_2nd_areachart = recycling_2nd.unstack(1)
recycling_2nd_areachart.columns = years
recycling_2nd_areachart = recycling_2nd_areachart.T

recycling_total_areachart = recycling_1st + recycling_2nd
recycling_total_areachart = recycling_total_areachart.unstack(1)
recycling_total_areachart.columns = years
recycling_total_areachart = recycling_total_areachart.T
landfill_total_areachart = landfill_1st + landfill_2nd
landfill_total_areachart = landfill_total_areachart.unstack(1)
landfill_total_areachart.columns = years
landfill_total_areachart = landfill_total_areachart.T
inflow_1st_areachart = inflow_1st_kg_mc_total.T

stock_europe_kg_mc_areachart = stock_europe_total_kg_mc_total.T

# Primary material = Inflow (material) - Recycled (material)
inflow_1st_recyclecorrected_areachart = inflow_1st_areachart - recycling_total_areachart
#%% Calculate total stock divided by 1st hand furniture and second hand furniture
stock_1st_kg_eu = pd.DataFrame(index=stock_1st_kg.index, columns={'stock_1st'})
stock_2nd_kg_eu = pd.DataFrame(index=stock_2nd_kg.index, columns={'stock_2nd'})
stock_1st_kg_eu['stock_1st']=stock_1st_kg['Central Europe']+stock_1st_kg['Western Europe']
stock_2nd_kg_eu['stock_2nd']=stock_2nd_kg['Central Europe']+stock_2nd_kg['Western Europe']
stock_kg_eu_total=pd.DataFrame(index=stock_1st_kg_eu.index, columns={'stock_1st', 'stock_2nd'})
stock_kg_eu_total['stock_1st']=stock_1st_kg_eu['stock_1st']
stock_kg_eu_total['stock_2nd']=stock_2nd_kg_eu['stock_2nd']
stock_kg_eu_total = stock_kg_eu_total.groupby(['t'], axis=0).sum()
#%% Convert all results in million kg to [Mt]- divide by 1000
stock_europe_kg_mc_areachart = stock_europe_kg_mc_areachart/1000
recycling_total_areachart = recycling_total_areachart/1000
landfill_total_areachart = landfill_total_areachart/1000
inflow_1st_areachart = inflow_1st_areachart/1000
stock_kg_eu_total = stock_kg_eu_total/1000
inflow_1st_recyclecorrected_areachart = inflow_1st_recyclecorrected_areachart/1000
stock_europe_kg_total_pc_areachart = stock_europe_kg_total_pc_areachart/1000
#%% Plotting results: changes headings and axis labels of graphs and name of saved files between scenarios
if scenario == 0:
    # Barchart stock europe per material category
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_kg_mc_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Total stock Europe per material category_Baseline')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Total stock Europe per material category_Baseline.png', bbox_inches = 'tight')
    
    # Barchart recycling total Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    recycling_total_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Recycling_total_EU_Baseline')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Recycling_Total_EU_Baseline.png', bbox_inches = 'tight')
    
    # Barchart landfill total Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    landfill_total_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('landfill_total_EU_Baseline')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('landfill_total_EU_Baseline.png', bbox_inches = 'tight')
    
    # Barchart inflow total Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    inflow_1st_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Inflow_1st_total_EU_Baseline')
    plt.xlabel('year')
    plt.ylabel('x million kg')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Inflow_1st_total_EU_Baseline.png', bbox_inches = 'tight')
    
    # Barchart stock total divided by 1st and 2nd hand furniture Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_kg_eu_total.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('stock_kg_eu_total_Baseline')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('stock_kg_eu_total_Baseline.png', bbox_inches = 'tight')
    
    # Required primary material total Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    inflow_1st_recyclecorrected_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Primary material_Baseline')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Primary material_Baseline', bbox_inches = 'tight')
   
    # Barchart total stock per product category
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_kg_total_pc_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Stock per product category_Baseline')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Stock_perproductcategory_Baseline.png', bbox_inches = 'tight')
    
    # Barchart office chair stock Europe 1st and 2nd hand Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_barchart_officechair.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Office chair stock 1st and 2nd hand_Baseline')
    plt.xlabel('year')
    plt.ylabel('x million units')
    plt.legend(bbox_to_anchor=(1.05,1))
    plt.savefig('Office_chair_stock_1and2_Baseline.png', bbox_inches = 'tight')
    
    # Barchart mattress stock Europe 1st and 2nd hand Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_barchart_mattress.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Mattress_stock 1st and 2nd hand_Baseline')
    plt.xlabel('year')
    plt.ylabel('x million units')
    plt.legend(bbox_to_anchor=(1.05,1))
    plt.savefig('Mattress_stock_1and2_Baseline.png', bbox_inches = 'tight')
elif scenario == 1:
    # Barchart stock europe per material category
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_kg_mc_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Total stock Europe per material category_studentscenario')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Total stock Europe per material category_studentscenario.png', bbox_inches = 'tight')
    
    # Barchart recycled material total student scenario
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    recycling_total_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Recycling_total_EU_studentscenario')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Recycling_Total_EU_studentscenario.png', bbox_inches = 'tight')
    
    # Barchart landfilled material total Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    landfill_total_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('landfill_total_EU_studentscenario')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('landfill_total_EU_studentscenario.png', bbox_inches = 'tight')
    # Barchart inflow Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    inflow_1st_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Inflow_1st_total_EU_studentscenario')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Inflow_1st_total_EU_studentscenario.png', bbox_inches = 'tight')
    
    #Barchart stock divided by 1st and 2nd hand furniture student scenario
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_kg_eu_total.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('stock_kg_eu_total_studentscenario')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('stock_kg_eu_total_studentscenario.png', bbox_inches = 'tight')
    # Barchart required primary material student scenario
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    inflow_1st_recyclecorrected_areachart.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Primary material_studentscenario')
    plt.xlabel('year')
    plt.ylabel('Mt')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Primary material_studentscenario', bbox_inches = 'tight')
    
    # Barchart office chair stock Europe 1st and 2nd hand Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_barchart_officechair.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Office chair stock 1st and 2nd hand_studentscenario')
    plt.xlabel('year')
    plt.ylabel('x million units')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Office_chair_stock_1and2_studentscenario.png', bbox_inches = 'tight')
    
    # Barchart mattress stock Europe 1st and 2nd hand Baseline
    colors = ["#FF0000", "#00FFFF", "#0000FF", "#0000A0", "#ADD8E6", "#800080", "#FFFF00", "#00FF00", "#FF00FF", "#C0C0C0", "#808080", "#000000", "#FFA500", "#800000", "#008000", "#808000"]
    stock_europe_barchart_mattress.plot(kind='area', stacked=True, color = colors)
    plt.xticks(years[0::5],rotation='vertical')
    plt.title('Mattress_stock 1st and 2nd hand_studentscenario')
    plt.xlabel('year')
    plt.ylabel('x million units')
    plt.legend(bbox_to_anchor=(1.05,1))
    #plt.savefig('Mattress_stock_1and2_studentscenario.png', bbox_inches = 'tight')
else:
    sys.exit("No (correct) scenario has been chosen. Please choose a scenario at the top of the program")
#%% Optional: total stock NOTE: this adds a column to the original dataframe, chaning its structure. Alternatively: a new dataframe can be made keeping the old dataframes intact.
stock_europe_kg_mc_areachart['Stock Total'] = stock_europe_kg_mc_areachart.sum(axis=1)
landfill_total_areachart['Total'] = landfill_total_areachart.sum(axis=1)
recycling_total_areachart['Total'] = recycling_total_areachart.sum(axis=1)
inflow_1st_recyclecorrected_areachart['Total'] = inflow_1st_recyclecorrected_areachart.sum(axis=1)
stock_europe_kg_total_pc_areachart['Total'] = stock_europe_kg_total_pc_areachart.sum(axis=1)
#
#"""-----------------------------------------> Part 4 Miscellaneous <--------------------------------------------------------"""
#%% piechart stock per product category
#stock_pc_2021 = stock_europe_kg_total_pc_areachart.loc[2021]
#stock_pc_2021 = stock_pc_2021.to_frame()
#floorspace_eu = floorspace.loc[:,'Western Europe']+floorspace.loc[:,'Central Europe']
#floorspace_eu = floorspace_eu.to_frame()
#floorspace_eu = floorspace_eu.groupby('t').sum()

#stock_pc_2021.plot.pie(subplots=True)
#plt.legend(bbox_to_anchor=(1.05,1))
#%%
#Optional: Total Population EU
pop_tot = pop_rur+pop_urb
pop_eu = pop_tot.loc[idx[:, 'Central Europe']]+pop_tot[idx[:, 'Western Europe']]